# -*- coding: utf-8 -*-

"""
 (c) 2017 - 2018 - Copyright Red Hat Inc

 Authors:
   Patrick Uiterwijk <puiterwijk@redhat.com>

   The Dream Team
     Pierre-Yves Chibon <pingou@pingoured.fr>
     Ralph Bean <rbean@redhat.com>

"""
from __future__ import print_function

import logging
import re
import os

try:
    from pdc_client import PDCClient
except ImportError:
    PDCClient = None


if "PAGURE_CONFIG" not in os.environ and os.path.exists(
    "/etc/pagure/pagure.cfg"
):
    os.environ["PAGURE_CONFIG"] = "/etc/pagure/pagure.cfg"

from pagure.config import config as pagure_config
from pagure.lib import model, get_user
from pagure.lib.git import is_forced_push
from pagure.lib.git_auth import GitAuthHelper, _read_file
from pagure.utils import is_repo_committer

_log = logging.getLogger(__name__)


class DistGitAuth(GitAuthHelper):
    """ A Fedora and CentOS dist-Git authentication module. """
    is_dynamic = True

    def __init__(self):
        super(DistGitAuth, self).__init__()

        self.is_debug = pagure_config.get("ACL_DEBUG", False)
        self.global_pr_only = pagure_config.get("PR_ONLY", False)
        self.block_unspecified = pagure_config.get("ACL_BLOCK_UNSPECIFIED",
                                                   True)
        self.blacklists = [re.compile(refre)
                           for refre
                           in pagure_config["BLACKLIST_RES"]]
        self.unspecified_blacklist = [
            re.compile(refre)
            for refre
            in pagure_config.get("UNSPECIFIED_BLACKLIST_RES", [])]
        self.rcm_branches = [re.compile(refre)
                             for refre
                             in pagure_config["RCM_BRANCHES"]]
        self.rcm_group = pagure_config.get("RCM_GROUP")
        self.supported_sigs = set(pagure_config.get("SUPPORTED_SIGS", []))
        self.sig_prefixes = pagure_config.get("SIG_PREFIXES", [])
        self.protected_namespaces = pagure_config.get(
            "ACL_PROTECTED_NAMESPACES", ["rpms"])

        pdc_url = pagure_config.get('PDC_URL')
        if pdc_url and PDCClient:
            self.pdcclient = PDCClient(pdc_url, develop=True)
        else:
            self.pdcclient = None

    def is_supported_branch(self, project, refname):
        """ Returns whether a specific branch is currently supported for Fedora

        This retrieves the information about EOL status from PDC, to prevent
        EOL branches being pushed to.
        """
        if not self.pdcclient:
            # No way to confirm this is a supported branch, not supported
            return None
        if not refname.startswith('refs/heads/'):
            # Anything outside of refs/heads is out-of-scope
            return None
        refname = refname[len('refs/heads/'):]

        namespace2pdctype = {
            'rpms': 'rpm',
            'modules': 'module',
            'container': 'container',
        }
        res = list(self.pdcclient.get_paged(
            self.pdcclient['component-branches'],
            global_component=project.name,
            type=namespace2pdctype[project.namespace],
            name=refname,
            fields=['active'],
        ))
        if len(res) == 0:
            # No status
            return None
        if len(res) != 1:
            # PDC couldn't make up its mind....
            # Should never happen, but just in case...
            raise ValueError("PDC was unable to make up its mind")
        return res[0]['active']

    def info(self, msg):
        """ Function to print information.

        Is a function to help with the test suite. """
        print(msg)

    def debug(self, msg):
        """ Function to print if in debug mode.

        Function to help in test. """
        if self.is_debug:
            self.info(msg)

    def check_acl(
        self,
        session,
        project,
        username,
        refname,
        pull_request,
        repodir,
        repotype,
        revfrom,
        revto,
        is_internal,
        **info
    ):
        """ Perform dynamic ACL checks.

        For arguments, see GitAuthHelper.check_acl."""

        # First, some rules regardless of user status
        # Only the "main" and "requests" repos are used
        if repotype not in ("main", "requests"):
            self.info("Repotype %s not in use" % repotype)
            return False

        # Do not allow branch deletions
        if revto == '0'*40:
            self.info("Branch deletion is not allowed")
            return False

        # Do not allow non-fast-forward pushes for anything but forks
        if not project.is_fork and is_forced_push(revfrom, revto, repodir):
            self.info("Forced pushes are not allowed")
            return False

        # If this is pushed through internal changes, approve.
        # This flag is only provided if it's either an update to the "side"
        # repos, or during creation/forking of a project.
        if is_internal:
            self.info("Internal push allowed")
            return True

        # Deploy Keys are not allowed on dist-git
        if username.startswith("deploykey_"):
            self.info("Deploy keys are disabled")
            return False

        # Now build up info about the user and project to make decisions
        # Determine the users' groups
        user = get_user(session, username)
        usergroups = set(user.groups)

        # Determine whether the user is a committer
        is_committer = is_repo_committer(project, username, session)

        # Determine whether the user is a SIG member
        user_sigs = self.supported_sigs & usergroups

        # Determine whether the user is RCM
        is_rcm = bool(self.rcm_group and self.rcm_group in usergroups)

        # If configured, print user info
        self.debug("Protected namespaces: %s" % self.protected_namespaces)
        self.debug("Blocking unspecified refs: %s" % self.block_unspecified)
        self.debug("Blacklists: %s" % self.blacklists)
        self.debug("User: %s" % user)
        self.debug("User groups: %s" % usergroups)
        self.debug('Committer: %s' % is_committer)
        self.debug('SIG memberships: %s' % user_sigs)
        self.debug('RCM: %s' % is_rcm)

        # We have data, start the actual ACL checking
        if (
            repotype == "main"
            and not project.is_fork
            and project.namespace in self.protected_namespaces
        ):
            # In the protected namespace, we want to make sure we don't
            # trample on blacklisted content.
            for entry in self.blacklists:
                if entry.match(refname):
                    self.info("Ref %s is blocked" % refname)
                    return False

            # Allow RCM push
            if is_rcm:
                for refre in self.rcm_branches:
                    if refre.match(refname):
                        # RCM on an RCM branch -> allow
                        self.debug("RCM push")
                        return True

            # Allow CentOS SIGs to push to e.g. c7-sig-core-$anything
            for signame in user_sigs:
                for sigprefix in self.sig_prefixes:
                    sigbranch = "%s-%s-" % (sigprefix, signame)
                    if refname.startswith(sigbranch):
                        self.debug("SIG push")
                        return True

            # For Fedora, allow supported branches
            is_supported = self.is_supported_branch(project, refname)
            if is_supported is False:
                self.info("Branch %s is unsupported" % refname)
                return False
            elif is_supported is True:
                self.debug("Branch %s is supported" % refname)
                return is_committer
            else:
                self.debug("No supported status available")

            if self.block_unspecified:
                self.info("Access to namespace %s is restricted"
                          % project.namespace)
                return False

            # Block second level blacklists
            for entry in self.unspecified_blacklist:
                if entry.match(refname):
                    self.info("Unspecified ref %s is blocked" % refname)
                    return False

            # For unspecified refs, they can push if they're a committer
            self.debug("Unspecified branch push")
            return is_committer

        # This is outside of the strongly protected namespaces
        if repotype == "main":
            # If this project has PRs only on, or PRs are globally enforced and
            # this is not a fork, only allow pushing if this is a PR merge.
            pr_only = project.settings.get("pull_request_access_only", False)
            if (
                pr_only
                or (self.global_pr_only and not project.is_fork)
                and pull_request is None
            ):
                self.info("A pull request is required for this branch")
                return False

        # This is an unprotected namespace, let's allow committers
        if is_committer:
            self.debug("Committer push to unprotected")
            return True

        # If all else fails, deny
        self.info("Fall-through deny")
        return False

    @classmethod
    def generate_acls(cls, project, group=None):
        """ Required function that doesn't do anything. """
        pass

    @classmethod
    def remove_acls(cls, session, project):
        """ Required function that is unused. """
        pass
